#!/usr/bin/env python3
"""Brython program simulating Einstein-Poincaré Synchronization."""

# © 2024 Jean-Pierre Messager
# License: MIT
# https://codeberg.org/jpm/pyweb-einstein

from random import random
from collections import defaultdict
from enum import Enum
from typing import Callable
from math import isclose
from browser import document, timer

# Model: Clocks and Synchronization classes
# NO WEB HERE ALLOWED!!!
# ONLY PHYSICS

class Clock:
    """Clock class."""
    randmin:float  = 10.00
    randmax:float  = 200.00
    interval:float = 0.01
    start:float    = 0.0
    def __init__(self):
        self._value = Clock.start
        self.__records = defaultdict(list)
    @classmethod
    def set_value(cls, interval):
        cls.interval = interval
    @property
    def value(self) -> float:
        return self._value
    @value.setter
    def value(self, new_value:float):
        self._value = new_value
    def tick(self):
        self.value += Clock.interval
    # Ensure consecutive values to be far away f.e.o
    # FIRST use of mutable default value in purpose! Hurra!
    # ??? marche pô...
    def random(self):
        self.value = random() * (Clock.randmax - Clock.randmin) + Clock.randmin
    def reset(self):
        self.value = 0.0
    def record(self, label:str) -> float:
        self.__records[label].append(self.value)
        return self.value
    def getrecords(self, label:str) -> list:
        return self.__records.get(label)

class SynchState(Enum):
    NOTDONE = "Not done or cleared"
    STEP1   = "Outgoing signal"
    STEP2   = "Return signal"
    DONE    = "Checked!"

class Synchronization:
    steps:int = 300
    def __init__(self, clock1, clock2, x1:float, x2:float):
        self.clock1, self.clock2 = clock1, clock2
        self.x1, self.x2 = x1, x2
        # TODO: is 0 allowed?
        self._dx = (x2 - x1)/Synchronization.steps
        self.reset()
    # This method could be Monkey Patched by
    # the view (i.e. the simulation)
    def go_state(self, state:SynchState):
        self.state = state
    def reset(self):
        self.state = SynchState.NOTDONE
        self.tA = self.tB = self.tpA = None
    def move(self):
        if self.state is SynchState.STEP1:
            self.x += self._dx
            # First signal received at c2?
            # then send signal back c2 -> c1
            if (self._dx > 0 and self.x >= self.x2) or \
               (self._dx < 0 and self.x <= self.x2):
                self.tB = self.clock2.record("tB")
                self.go_state(SynchState.STEP2)
        # Second signal received at c1?
        if self.state is SynchState.STEP2:
            self.x -= self._dx
            if self.x <= self.x1 if self._dx > 0 else \
               self.x >= self.x1:
                self.tpA = self.clock1.record("t'A")
                self.go_state(SynchState.DONE)
                self.x = self.x1
    def start(self): # Send signal c1 -> c2
        self.go_state(SynchState.NOTDONE)
        self.tA = self.clock1.record("tA")
        self.tB = self.tpA = None
        #self.x = 0.0
        self.x = self.x1
        self.go_state(SynchState.STEP1)
    def isAsynchB(self):
        #if self.state is not SynchState.DONE:
        #    return None
        # Compute offsets
        self.Aoffset = self.tB - (self.tpA - self.tA)/2 - self.tA
        self.Boffset = (self.tpA + self.tA)/2 - self.tB
        # Normalize to 0
        self.Aoffset = 0.0 \
                if isclose(self.Aoffset, 0.0, rel_tol=Clock.interval) else self.Aoffset
        self.Boffset = 0.0 \
                if isclose(self.Boffset, 0.0, rel_tol=Clock.interval) else self.Boffset
        # Einstein formula (**)
        return isclose(self.tB - self.tA, self.tpA - self.tB, \
                rel_tol=Clock.interval)

# VIEWS: Simulation class and decoration on the Synchronization model.
# Monkey Patching is acceptable, when used properly. We are all consenting
# adults here, right?
# Quite a mess (side effects are a plague, Web is a huge mess).
#  "Views know the model, the model does not known anything about
#   the views."
# Feel free to improve :-)

def add_simulation_hooks(simulation) -> Callable:
    def _(cls:Synchronization) -> Synchronization:
        def ND_S1(self):
            simulation.doc["tA"].value = f"{self.tA:.2f}"
            simulation.doc["tB"].value = ""
            simulation.doc["tpA"].value = ""
            simulation.doc["AsynchB"].disabled = True
            simulation.doc["BsynchA"].disabled = True
            simulation.doc["resAsynchB"].value = "👷"
            simulation.doc["setA"].disabled = True
            simulation.doc["setB"].disabled = True
        def S1_S2(self):
            simulation.doc["tBminustA"].value = \
                    f"{simulation._sync.tB - simulation._sync.tA:.2f}"
            simulation.doc["tB"].value = f"{self.tB:.2f}"
        def S2_D(self):
            simulation.doc["tpA"].value = f"{self.tpA:.2f}"
            simulation.doc["AsynchB"].disabled = False
            simulation.doc["BsynchA"].disabled = False
            simulation.doc["tpAminustB"].value = \
                    f"{simulation._sync.tpA - simulation._sync.tB:.2f}"
            if simulation._sync.isAsynchB():
                simulation.doc["resAsynchB"].value = "✅"
            else:
                simulation.doc["resAsynchB"].value = "❌"
            if simulation._sync.x1 < simulation._sync.x2:
                simulation.doc["Aoffset"].value = \
                        f"{simulation._sync.Aoffset:+.2f}"
                simulation.doc["Boffset"].value = \
                        f"{simulation._sync.Boffset:+.2f}"
            else:
                simulation.doc["Aoffset"].value = \
                        f"{simulation._sync.Boffset:+.2f}"
                simulation.doc["Boffset"].value = \
                        f"{simulation._sync.Aoffset:+.2f}"
            simulation.doc["doAoffset"].disabled = False
            simulation.doc["doBoffset"].disabled = False
            simulation.doc["setA"].disabled = False
            simulation.doc["setB"].disabled = False
        def D_ND(self):
            for id in ("tA", "tB", "tpA"):
                simulation.doc[id].value = ""
            simulation.doc["x"].value = ""
            simulation.doc["resAsynchB"].value = "❔"
            simulation.doc["AsynchB"].disabled = False
            simulation.doc["BsynchA"].disabled = False
        def invalid(self):
            print("Invalid transition?!")
        hooks = {
                (SynchState.NOTDONE, SynchState.STEP1):ND_S1,
                (SynchState.STEP1, SynchState.STEP2):S1_S2,
                (SynchState.STEP2, SynchState.DONE):S2_D,
                (SynchState.DONE, SynchState.NOTDONE):D_ND,
                }
        orig_go_state = cls.go_state
        def new_go_state(self, state:SynchState):
            old, new = self.state, state
            hooks.get((old, new), invalid)(self)
            orig_go_state(self, state)
            simulation.doc["simustate"].value = new.value
        cls.go_state = new_go_state
        return cls
    return _

class Simulation:
    """Simulation class."""
    timer           = None
    frozen:bool     = True
    interval:int    = 10
    play_label:str  = b'\xe2\x96\xb6\xef\xb8\x8f'.decode("utf-8")
    pause_label:str = b'\xe2\x8f\xb8\xef\xb8\x8f'.decode("utf-8")
    def __init__(self, doc, xA, xB, **clocks):
        (self.idA, self.clockA), \
        (self.idB, self.clockB) = clocks.items()
        self.xA, self.xB = xA, xB
        self.eticks = 0.0
        self.doc = doc
        for id in ("tA", "tB", "tpA"):
            self.doc[id].value = ""
        self.doc["random"].bind("click", self.reset_random)
        self.doc["freeze"].bind("click", self.freeze_unfreeze)
        self.doc["freeze"].value = Simulation.play_label
        self.doc["AsynchB"].bind("click", self.syncAonB)
        self.doc["BsynchA"].bind("click", self.syncBonA)
        self.doc["pulse"].html = "⭐"
        self.doc["pulse"].html = "💫"
        self.doc["pulse"].html = "💛"
        # ⚽🎱🍆🦟🐇🐢 🦋 🐌 🚆 🇸 🇵 🍉 🥧
        self.doc["pulse"].html = "🍉"
        self.doc["pulse"].html = "🥧"
        self.doc["pulse"].html = "🌟"
        #self.doc["pulse"].html = '<img src="pulse_red.png"/>'
        #@add_simulation_hooks(self)
        # class SynchronizationHooked(Synchronization):
        #    pass
        #self._sync = SynchronizationHooked(self.clockA, self.clockB, \
        #                   self.xA, self.xB)
        # to be computed again during animation as the window
        # can be resized
        #self._pulsewidth = self.doc["pulse"].offsetWidth
        #self._spacewidth = self.doc["interspace"].offsetWidth - self._pulsewidth
        self.reset_random(None)
        self.doc["simustate"].value = SynchState.NOTDONE.value
        self.doc["stopsimu"].value = "💣 Clear"
        self.doc["stopsimu"].bind('click', self.reset_sync)
        self.doc["clocksstate"].value = "Frozen"
        self.doc["AsynchB"].disabled = True
        self.doc["BsynchA"].disabled = True
        # ✅❌👍👎❔👷
        self.doc["resAsynchB"].value = "❔"
        self.doc["doAoffset"].bind("click", \
                (lambda ev: self.offset(self._sync.clock1, self._sync.Aoffset)))
        self.doc["doBoffset"].bind("click", \
                (lambda ev: self.offset(self._sync.clock2, self._sync.Boffset)))
        self.doc["setA"].bind("click", \
                (lambda ev: self.set_clock(self.clockA, "SetClockA")))
        self.doc["setB"].bind("click", \
                (lambda ev: self.set_clock(self.clockB, "SetClockB")))
    def reset_sync(self, ev=None):
        if not hasattr(self, "_sync"): # no Synch set yet
            return
        self._sync.reset()
        self.update_pulse()
        self.doc["simustate"].value = self._sync.state.value
        self.doc["pulse"].style.transform = "translate(0px,0)"
        self.doc["x"].value = ""
        self.doc["doAoffset"].disabled = True
        self.doc["doBoffset"].disabled = True
        self.doc["Aoffset"].value = ""
        self.doc["Boffset"].value = ""
        self.doc["tBminustA"].value = ""
        self.doc["tpAminustB"].value = ""
        self.doc["resAsynchB"].value = "❔"
        if not self.frozen:
            self.doc["AsynchB"].disabled = False
            self.doc["BsynchA"].disabled = False
    def update_pulse(self):
        self._pulsewidth = self.doc["pulse"].offsetWidth
        self._spacewidth = self.doc["interspace"].offsetWidth - self._pulsewidth
        if not hasattr(self._sync, "x"):
            return
        scr_x = (self._sync.x - self.xA)/(self.xB - self.xA)*self._spacewidth
        self.doc["pulse"].style.transform = f"translate({scr_x}px,0)"
        self.doc["x"].value = f"{self._sync.x:.2f}"
    def run(self, is_reset=False):
        # Internal counter (unphysical)
        if not is_reset:
            self.eticks += Clock.interval
        self.doc["etime"].value = f"{self.eticks:.2f}"
        # Run clocks
        for id, clock in (self.idA, self.clockA), (self.idB, self.clockB):
            self.doc[id].text = f"{clock.value:.2f}"
            if not is_reset:
                clock.tick()
        self.doc["simuClocksInSync"].value = "✅" if \
                isclose(self.clockA.value, self.clockB.value, Clock.interval) \
                else "❌"
        # Run synch if appropriate
        if not hasattr(self, "_sync"): # no Synch set yet
            return
        if self._sync.state == SynchState.STEP1 or \
           self._sync.state == SynchState.STEP2:
            if not is_reset:
                self._sync.move()
            self.update_pulse()
    def freeze_unfreeze(self, ev=None):
        if self.timer is not None:
            timer.clear_interval(self.timer)
            self.timer = None
            self.frozen = True
            self.doc["freeze"].value = Simulation.play_label
            self.doc["clocksstate"].value = "Frozen"
            self.doc["AsynchB"].disabled = True
            self.doc["BsynchA"].disabled = True
        else:
            self.timer = timer.set_interval(self.run, Simulation.interval)
            self.frozen = False
            self.doc["freeze"].value = Simulation.pause_label
            self.doc["clocksstate"].value = "Running"
            self.doc["AsynchB"].disabled = False
            self.doc["BsynchA"].disabled = False
    def reset(self, ev=None):
        self.clockA.reset()
        self.clockB.reset()
        self.eticks = 0.0
        if self.frozen:
            self.run(is_reset=True)
        self.reset_sync()
    def reset_random(self, ev=None):
        self.clockA.random()
        self.clockB.random()
        self.eticks = 0.0
        if self.frozen:
            self.run(is_reset=True)
        self.reset_sync()
    def set_clock(self, clock, field):
        try:
            clock.value = float(self.doc[field].value.replace(",","."))
            if self.frozen:
                self.run(is_reset=True)
            self.reset_sync()
        except TypeError as e:
            print(f"Error: {e.args[0]}")
    # TODO: factorize int lambdas :-) Hello, Guido!
    def syncAonB(self, ev=None):
        if self.frozen:
            return
        @add_simulation_hooks(self)
        class SynchronizationHooked(Synchronization):
            pass
        self._sync = SynchronizationHooked(self.clockA, self.clockB, \
                           self.xA, self.xB)
        if not self.frozen:
            self._sync.start()
    def syncBonA(self, ev=None):
        if self.frozen:
            return
        @add_simulation_hooks(self)
        class SynchronizationHooked(Synchronization):
            pass
        self._sync = SynchronizationHooked(self.clockB, self.clockA, \
                           self.xB, self.xA)
        self._sync.start()
    def offset(self, clock:Clock, offset:float):
        clock.value += offset
        self.eticks = 0.0
        if self.frozen:
            self.run(is_reset=True)
        self.reset_sync()

# Controller :-)
# (really: the brower is the controller)

if __name__ == "__main__":
    simulation = Simulation(document, 0.0, 100.0, \
                        valueA=Clock(), valueB=Clock())

